#!/usr/bin/perl -w

# $Id: guardcat.pl,v 12.28 2023/01/25 17:54:57 bscott Exp bscott $

# This is free and unencumbered software released into the public domain.
# It is provided "AS IS", without warranty or liability of any kind.
# See the UNLICENSE file or http://unlicense.org/ for full license terms.

########################################################################
# imports

use strict;
use warnings;
use v5.020; # signatures, copy-on-write strings
use English qw( -no_match_vars );
use POSIX qw(strftime ceil);
use File::Basename;
use Sys::Hostname;
use Sys::Syslog qw(:standard :macros);
use Scalar::Util qw( looks_like_number );
no warnings qw( experimental::signatures experimental::try );
use feature qw( signatures try );

########################################################################
# global constants

# Program Identification
# These are variables to Perl, but constant in how they are used.
# I put them in Perl variables simply for convenient interpolation.
# The version and date are set during the build/release process.
# I'd prefer to get them from RCS-style keywords but git doesn't do that.
# The "subver.pl" script that does the replacements is very sensitive
# to minor variations in source code formatting, hence the warnings.
my $PROGNAME = 'guardcat';
# *** NEXT TWO LINES ARE AUTOMATICALLY EDITED -- DO NOT TOUCH **
my $VERSION  = '1.12.28';
my $PROGDATE = '2023-01-25';
# *** PREVIOUS TWO LINES ARE AUTOMATICALLY EDITED -- DO NOT TOUCH **

use constant {

# defaults
CFGFILE_DEFAULT    => '/etc/guardcat.conf',
POLLINT_DEFAULT    => 10,
LOGINT_DEFAULT     => 6,
ALERT_WAIT_DEFAULT => 36000,
RESET_WAIT_DEFAULT => 3600,

# pid = include process ID in log msg
# ndelay = open socket as soon as openlog() is called, not first syslog()
SYSLOG_OPTS => 'pid,ndelay',

# timestamp format YYYY-MM-DD HH:MM:SS
TIME_FORMAT => '%F %T',

# We start each cycle by setting every channel reading to -1 (negative one).
# If anything fails to read, we just keep that value.
# That is obviously-out-of-range yet still plottable.
# Might be a transient failure, and will work the next time around.
BAD_READ => -1,

# Kernel hwmon API gives temperatures in millidegrees.  We multiply
# by this to turn them into degrees (with fractional part).
MILLIDEGREES =>  0.001,

# input channel types
IN_FILE => 1,	# read directly from a file
IN_PROG => 2,	# read from the output of an external program
IN_MIN  => 3,   # aggregated, lowest of a set of other inputs
IN_MAX  => 4,   # aggregated, highest of a set of other inputs
IN_AVG  => 5,   # aggregated, mean of a set of other inputs

# Error behavior, given as arguments to some subs.
DIE_ON_ERROR      => 1,
CONTINUE_ON_ERROR => 0,

# Aggregation type for inputs_math_agg()
AGG_STRAIGHT    => 1, # aggregate the cooked (adjusted raw) values
AGG_NORMALIZED  => 2, # aggregate the normalized values

# Argument to poll_main()
POLL_SLOW_ONLY => 1, # TRUE, poll slow inputs only

# I often find words clearer than the Perl punctuation soup,
# especially when they are standing alone and look like a typo.
TAB	=> qq{\t}, # ASCII tab, used for tabular data
EMPTY	=> q{},   # zero-length string, null string
SPACE   => q{ },  # ASCII space
COLON   => q{:},
SEMI    => q{;},
NEWLINE	=> qq{\n}, # end-of-line, linefeed

# open() file I/O modes
READ_FROM => '<',  # read from an existing file
WRITE_TO  => '>',  # write to an existing file, truncating it
APPEND_TO => '>>', # append to an existing file
PIPE_TO   => '|-', # pipe to another program (we write to it)
PIPE_FROM => '-|', # pipe from another program (we read from it)

# boolean values
FALSE => 0,
TRUE  => 1,

# Used to mark the main loop that runs "forever".
FOREVER => 1,

# When given a single space, split() splits on any whitespace.
# This makes that explicit in the code.
SPLIT_ON_WHITESPACE => q{ },

# obscenely long line to mark poll cycles in debug output
POLLMARK => 'poll: ---------- sleep -------------------------------------------------- ',

}; # constant

########################################################################
# structures

# While the config file uses a variety of directives for inputs, aggregates,
# outputs, alarms, etc., here in the program, we keep most of it in two
# overall structures, %inputs and @outputs, each containing one-or-more of the
# corresponding Structs below.  Each output channel references an
# input/aggregate channel.  Input channels are not required to have an output
# channel (they may just be for logging, for example).

# an input channel (sensor) with related parameters for alarms, etc.
# or an aggregate channel, which is much the same thing
use Class::Struct Input => {
	type  => '$', # type of input: IN_FILE, IN_PROG, IN_MAX, etc.
	min   => '$', # defines bottom of input range for normalization
	max   => '$', # defines top    of input range for normalization
	mul   => '$', # multiplier, if any - applied before offset
	ofs   => '$', # offset, if any - applied after multiplier
	agg   => '$', # boolean, is this an aggregate input?
	round => '$', # boolean, round input value?  (looks neater in output)
	almin => '$', # below this, we alert
	almax => '$', # above this, we alert
	therm => '$', # boolean, is this a thermal reading (temperature)
	slow  => '$', # boolean, only sample when logging?
	path  => '$', # path to file/prog/etc
	args  => '@', # IN_PROG:arguments to program; IN_agg:other inputs
	raw   => '$', # raw value read from file/prog
	cook  => '$', # adjusted value, with scaling, Fahrenheit, etc.
	norm  => '$', # normalized value, from 0 to 1
	used  => '$', # boolean, set by log/alarm/output, used for cfg chk
	};

# an output channel (fan) with control parameters
use Class::Struct Output => {
	ichan => '$', # name of input channel that controls this output
	path  => '$', # file we write to, to set target value
	min   => '$', # defines bottom of control range for normalization
	max   => '$', # defines top    of control range for normalization
	norm  => '$', # normalized target value, from 0 to 1
	want  => '$', # actual target value to write to path
	floor => '$', # minimum allowed actual target value, overriding .want
	};

# Here a "Tweak" is a value we write to a file at startup or shutdown.
# Both @inits and @exits contain Tweaks.
use Class::Struct Tweak => {
	path  => '$', # file we write to
	want  => '$', # value we write
	};

########################################################################
# global vars

# ----------------------------------------------------------------------
# environment, determined automatically

my $host;     # hostname

# ----------------------------------------------------------------------
# general program state

# have we made any adjustments to outputs yet?
# if not, we don't try to restore things on program exit
my $adjusted = FALSE;

# epoch time we last alerted on out-of-bounds, if any
# used to know when to repeat the alerts (see $alert_wait)
my $alert_last = 0;

# epoch time a fault was last seen
# used to know when to consider trouble cleared (see $reset_wait)
my $fault_last = 0;

# ----------------------------------------------------------------------
# per-poll-cycle state

# log cycle, counts polls between data-log-writes
# set by log_next(), checked by log_decide()
my $logcyc = 0;

# log timestamp for current cycle
# global so alert mail and data log entries can be correlated by timestamp
my $stamp;

# Indicates log file should be re-opened between poll cycles.
# Set by sig_hup() in response to SIGHUP, checked in poll_cycle()
my $log_reopen = FALSE;

# Indicates the poll cycle should be immediately repeated, without logging
# or sleeping.  Used to ensure all sensors are read (including slow ones)
# after some kind of event occurs mid-cycle.
my $poll_repeat = FALSE;

# ----------------------------------------------------------------------
# config, loaded from config files/commands

# print/log gobs of debugging info?
my $debug = FALSE;

# Have we hit a config error and reported it?
# Used by cfg_load_argv() and cfg_load_file().
# Once the first error is reported, we don't keep reporting the same
# error as the exception propagates back up the stack.
my $cfgerr = FALSE;

# Convert temperature values to Fahrenheit?
# The omitted leading letter 'A' is intentional.  :-)
my $merica = FALSE;

# email this address to alert of out-of-bounds sensors
my $mail_to = FALSE;

# send out-of-bounds alerts to all users with wall(1)?
my $wall = FALSE;

# Suppress warn() to stderr?
# We always do syslog, and we always write errors to stderr.
# This suppresses dupe logs when systemd is capturing stderr to logs.
my $quiet = FALSE;

# Input channels (sensors) defined in config file.
my %inputs = ();

# Input names, sorted, kept in order of definition.
# Order is important for dependencies of aggregate inputs.
# Also allows reports and such to be in the same order they were defined.
my @sinputs = ();

# Output channels (fans), and controls thereof.
my @outputs = ();

# Tweaks to write at program startup
my @inits = ();

# Tweaks to write at program shutdown
my @exits = ();

# every X seconds, poll sensors (read), set outputs (write)
my $pollint = POLLINT_DEFAULT;

# Every X poll cycles, write sensor readings to data log.
my $logint = LOGINT_DEFAULT;

# Names of input channels (sensors) to write to data log, in order.
my @logchans = ();

# Filename to write sensor data to (not syslog).
my $logfile = EMPTY;

# File handle for data log (not syslog), if open.
# If not defined, no logging is done.
my $log;

# When a value is out-of-bounds, every X seconds, send an alert.
my $alert_wait = ALERT_WAIT_DEFAULT;

# When all values are back in-bounds for X seconds, consider trouble cleared.
my $reset_wait = RESET_WAIT_DEFAULT;

# Was at least one alarm declared in the config file?
# Used to sanity check the config after load.
my $at_least_one_alm = FALSE;

########################################################################
# main program

die 'do not run an unreleased file' if ($VERSION  eq 'UNRELEASED')
                                    or ($PROGDATE eq 'UNDATED');

# Program normally repeats main loop forever.
# Normal program exit is done by signals and signal handlers (see below).
# There is an END block to handle cleanup.
# Top-level exception handler here, mostly so errors can be copied to syslog.

try {
	main_init();
	syslog(LOG_INFO, 'running');

	# MAIN LOOP
	while (FOREVER) {
		poll_cycle();
		}

	die 'infinite loop exited?!?';
	}
catch ($e) {
	if ($debug) {
		dmsg("!!!!! $e");
		cfg_dump();
		}
	emsg($e);
	emsg('Exiting on fatal error' . NEWLINE);
	exit 1;
	}

########################################################################
# major subroutines

# ----------------------------------------------------------------------
sub poll_cycle() {
# This is the body of the main program loop.
# Each call to this sub is one poll cycle.
# This calls on the workers, then does the sleep for the poll interval.	

$poll_repeat = FALSE; # reset at start of cycle
$logcyc++;
$stamp = strftime(TIME_FORMAT, localtime);
dmsg("poll: logcyc=$logcyc stamp=$stamp", __LINE__) if $debug;
poll_clear();
poll_main();
bounds_main();
return if $poll_repeat; # return to main loop, which calls us again
log_maybe();
log_reopen() if $log_reopen; # in case of SIGHUP during poll cycle

# sleep until the next poll cycle (or interrupt)
dmsg(POLLMARK, __LINE__) if $debug;
sleep $pollint;
log_reopen() if $log_reopen; # in case of SIGHUP during sleep

} # poll_cycle

# ----------------------------------------------------------------------
sub poll_main($only_slow = FALSE) {
# Do the inputs and outputs part of polling.
# This is broken out separately so that bounds_main() can call it again with
# POLL_SLOW_ONLY if an alert occurs during a log_decide()=FALSE cycle.
# See inputs_read() for more on $only_slow.

inputs_read($only_slow);

# Slow, fast, or both, the rest are always the same.
# Calling them twice in one cycle is also fine.
# The math will generally be the same or work out better (with more input).
# The outputs may thus may be more correct.  Writing them is good, then.
# If not, writing the same output twice is fine.
inputs_math();
outputs_math();
outputs_write();

} # poll_main

# ----------------------------------------------------------------------
sub main_init() {
# Assorted global initialization of main program.  In a sub for neatness.

$host = hostname;

# initialize syslog early, so we can log startup problems
openlog($PROGNAME, SYSLOG_OPTS, LOG_DAEMON);
syslog(LOG_NOTICE, "$PROGNAME version $VERSION ($PROGDATE) starting");

# likewise the handler that copies warnings to syslog
$SIG{__WARN__} = \&sig_warn;

# handle "normal shutdown" signals (daemon exit)
$SIG{INT} = $SIG{TERM} = $SIG{QUIT} = \&sig_end;

# signals for serious problems
$SIG{ABRT} = $SIG{SEGV} = $SIG{BUS} = $SIG{ILL} = $SIG{FPE} =
    $SIG{SYS} = \&sig_crash;

cfg_init(); # load, check, and initialize runtime configuration

# write values from config "init" directives
tweaks_write(DIE_ON_ERROR, @inits);

# handle "operational" signals - changes in running state, not shutdown
$SIG{USR1} = \&sig_usr1;
$SIG{USR2} = \&sig_usr2;
$SIG{HUP}  = \&sig_hup;

} # main_init

########################################################################
# signal handlers

# Unlike regular subroutines, I do not specify a call signature here.
# The Perl documentation is sketchy on what arguments might be passed.
# Pretty sure it's just the signal, but no promises are made, so let's be
# flexible.

# ----------------------------------------------------------------------
sub sig_hup {
# On SIGHUP, reopen the log file with next poll cycle.  Intended for use after
# log rotation.
#
# We just set a flag, and check the flag in poll_cycle(), to reduce problems
# with races if a signal comes in at an inopportune time.

dmsg('received SIGHUP', __LINE__) if $debug;
return unless $log;
$log_reopen = TRUE;

} # sig_hup

# ----------------------------------------------------------------------
sub sig_usr1 {
# Turn SIGUSR1 into an immediate read-sensors-and-log-results.
# If debugging, also dumps current config (including sensors and targets).
#
# Timing scenarios:
# 1. If we were sleeping, we wake up and then do exactly what we want.
# 2. If we were in the middle of a poll cycle, there are some potential races.
# 2a. There's a window mid-cycle where it might not read a slow input,
#     (because at the time it wasn't going to log, and the signal came in
#     after).  It would thus log BAD_READ (-1) for the slow sensors.
#     We defuse this by setting poll_repeat as well as log_next.
#     There's still a tiny window (between the $poll_repeat check and the
#     actual logging), but that's not worth worrying about.  I hope.
# 2b. There's a tiny window where it might not log immediately, but instead
#     with the next poll cycle (if the signal comes in after log_decide()).
#     This also seems like it's not worth worrying about.  I hope.
# 2c. SIGUSR1 is mainly intended for interactive use when someone is
#     watching the log file, so they can always send the signal again.
#
# I looked into doing this with poll_main(POLL_SLOW_ONLY) but I still needed a
# global flag and logic to handle it and it added more timing issues, so
# $poll_repeat is simpler for this.  Unlike bounds_main().

if ($debug) {
	dmsg('received SIGUSR1', __LINE__);
	cfg_dump();
	}

# if we don't have a log file, we can't do an immediate log to it.
return unless $log;

# If we weren't going to log data at the end of this log cycle...
if (not log_decide()) {
	# ... now we are!
	log_next();
	# If the signal came in mid-cycle, repeat the cycle, so
	# that any slow sensors get their data read and logged, too.
	# If the signal came in during sleep, poll_repeat gets reset anyway.
	$poll_repeat = TRUE;
	}

# If we were going to log data anyway, we don't have to do anything
# to satisfy the SIGUSR1 request.

} # sig_usr1

# ----------------------------------------------------------------------
sub sig_usr2 {
# Toggle debugging with SIGUSR2.

# dmsg is called when/after debugging is (still) enabled
if ($debug) {
	dmsg('debugging disabled by SIGUSR2', __LINE__);
	$debug = 0;
	}
else {
	$debug = 1;
	dmsg('debugging enabled by SIGUSR2', __LINE__);
	}

} # sig_usr2

# ----------------------------------------------------------------------
sub sig_end {
# Called for signals that are normally used to shutdown a daemon, like SIGINT
# and SIGTERM.
#
# We redirect them into a Perl exit(), which should result in the END block
# being run, which should in turn call our cleanup code.

# Ignore further shutdown signals; we are doing it as fast as we can.
# Also ignore operational signals (USR1/log and HUP/reopen).
$SIG{INT} = $SIG{TERM} = $SIG{QUIT} = $SIG{USR1} = $SIG{USR2} = $SIG{HUP} = 'IGNORE';
# Keep __WARN__ and __DIE__ handlers.  If something goes wrong during
# routine shutdown, we certainly want to try to log that.

my $msg = "Exiting on signal $_[0]";
syslog(LOG_NOTICE, $msg);
say {*STDERR} ("$PROGNAME: $msg");

exit 0; # exit program (to END routines), indicating success

} # sig_end

# ----------------------------------------------------------------------
sub sig_crash {
# Called for signals that indicate things have gone very wrong for the Perl
# runtime, like SIGSEGV (segmentation fault).
#
# Of course, if the runtime is that badly broken, there's a good chance it
# will fall apart trying to run this handler, but it's worth a try.  Assuming
# the runtime is sane, the die() below should still result in the END block
# being run, so hopefully we still get proper cleanup.

# Ignore routine shutdown signals, since we are dying anyway.
# Ignore operational signals (USR1/log and HUP/reopen).
$SIG{INT} = $SIG{TERM} = $SIG{QUIT} = $SIG{USR1} = $SIG{USR2} = $SIG{HUP} = 'IGNORE';
# If we get another such signal (double fault), just abort normally.
# Disable the die() handler so our die below doesn't log twice.
$SIG{ABRT} = $SIG{SEGV} = $SIG{BUS} = $SIG{ILL} = $SIG{FPE} =
    $SIG{SYS} = $SIG{__DIE__} = 'DEFAULT';

# Top-level exception handler should catch this, if that's working.
# If not, hopefully Perl will do a standard die() (print msg and exit).
die "Received fatal signal $_[0]";

} # sig_crash

# ----------------------------------------------------------------------
sub sig_warn {
# Trap warnings and send them to syslog as well as stderr.

# suppress this handler in case of double-fault
$SIG{__WARN__} = 'DEFAULT';

syslog(LOG_WARNING, (join SPACE, @_) );

# this is mostly what warn() normally does
print {*STDERR} ("$PROGNAME: ", @_) unless $quiet;

# restore for future warnings
$SIG{__WARN__} = \&sig_warn;

}

# ----------------------------------------------------------------------
END {
# This handles cleanup when the program exits normally or aborts.
# (Technically an END block is not a signal handler, but close enough.)

dmsg('entering END block', __LINE__) if $debug;

# We don't try to do anything with signal handlers here.  All the code paths
# that lead to exiting Perl should already have set them appropriately.

# Write out any "exit" values declared in the config.  These typically set
# fans to full speed, so the system doesn't cook itself if this program
# crashes.  Only do this if we made adjustments in the first place.
# Treat I/O errors as warnings only, and blindly try to continue, because
# we're shutting down (or crashing) and there's not much else to do.
tweaks_write(CONTINUE_ON_ERROR, @exits) if $adjusted;

dmsg('leaving END block', __LINE__) if $debug;

} # block END

########################################################################
# output subroutines

# Anything involved in writing out a value for fan/hardware control.
# These effect things in the real world.  Everything else is just info.
# (I suppose technically logs are an output, too, but not this kind.)

# ----------------------------------------------------------------------
sub outputs_math() {
# Determine target values for each output by examining the input control
# definitions and performing appropriate calculations.

# for each of the outputs
OUT: foreach my $out (@outputs) {

	my $ichan = $out->ichan;
	my $min = $out->min;
	my $max = $out->max;

	# The normalized value from the controlling input becomes
	# the target normalized value for this output.
	my $normed = $inputs{$ichan}->norm;

	# should never happen
	die 'bogus normed' if (not defined $normed) or ($normed == BAD_READ);

	$out->norm($normed);

	# set the output's target value based on the normalized target
	my $want = denorm($normed, $min, $max);
	dmsg("mathout: chan=$ichan min=$min max=$max want=$want", __LINE__)
		if $debug;
	$out->want($want);

	}

} # outputs_math

# ----------------------------------------------------------------------
sub outputs_write() {
# Write target value (desired fan speed) to each output channel

CHAN: foreach my $out (@outputs) {

	my $path  = $out->path;
	my $want  = $out->want;
	my $floor = $out->floor;

	die 'out.want undef' unless defined $want; # should never happen

	$want = $floor if defined $floor and $want < $floor;

	dmsg("output: $want >>> $path", __LINE__) if $debug;

	my $opened = open(my $outfile, WRITE_TO, $path);
	if (not $opened) {
		warn "output: $path: open: $!";
		next CHAN;
		}

	# only after we successfully open an output, do we process @exits
	$adjusted = TRUE;

	say {$outfile} $want or warn "output: $path: write: $!";
	close $outfile or warn "output: $path: close: $!";

	}

} # outputs_write

# ----------------------------------------------------------------------
sub tweaks_write ($errdie, @tweaklist) {
# Given a list of Tweak structures, write the value to the path for each.
# If $errdie is TRUE, any problems will die(); else, simply warn.
# Used to process @inits and @exits at startup and shutdown, respectively.

dmsg('tweaks_write: entering', __LINE__);

# $prob is a ref to the function we call if there is a problem.
# $prob = die() if $errdie, else $prob = warn()
my $prob = $errdie ? \&CORE::die : \&CORE::warn ;

CHAN: foreach my $tweak (@tweaklist) {
	my $path = $tweak->path();
	my $want = $tweak->want();
	my $opened = open(my $fh, WRITE_TO, $path);
	if (not $opened) {
		$prob->("tweak: $path: open: $!");
		next CHAN;
		}
	dmsg("tweak: $want >>> $path", __LINE__);
	say {$fh} $want or $prob->("tweak: $path: write: $!");
	close $fh or $prob->("tweak: $path: close: $!");
	# If $tweaklist was @inits we want to remember we changed something.
	# (If $tweaklist was @exits, $adjusted was what got us here, but OK.)
	$adjusted = TRUE;
	}

dmsg('tweaks_write: exiting', __LINE__);

}

########################################################################
# input subroutines

# These are all about reading data inputs (sensor values).
# (Technically the config file is an "input", but that's different.)

# ----------------------------------------------------------------------
sub inputs_math() {
# Do math for each input channel, as appropriate (direct or aggregate).

CHAN: foreach my $chan (@sinputs) {

	if ($inputs{$chan}->agg) {
		inputs_math_agg($chan, AGG_STRAIGHT);
		inputs_math_agg($chan, AGG_NORMALIZED);
		}
	else {
		inputs_math_direct($chan);
		}

	# rounding is the only option that applies to agg and direct
	if ($inputs{$chan}->round) {
		my $val = $inputs{$chan}->cook;
		# Round to one fractional decimal place, if needed.
		# If already an integer, don't add the ".0".
		if ( ($val != BAD_READ) and ($val != int($val)) ) {
			$val = sprintf('%.1f', $val) ;
			$inputs{$chan}->cook($val);
			}
		}

	} # CHAN

} # inputs_math

# ----------------------------------------------------------------------
sub inputs_math_direct($chan) {
# Do math on the most-recent readings for a given direct input channel.
# Apply scaling and offset.  Temperature conversion.  Normalize.

my $val = $inputs{$chan}->raw;
my $min = $inputs{$chan}->min;
my $max = $inputs{$chan}->max;
my $mul = $inputs{$chan}->mul;
my $ofs = $inputs{$chan}->ofs;

# If we failed to get a good reading, keep the BAD_READ, and use the max
# for the normalized.  That will keep the plot looking correct, and set
# fans to highest.
if ($val == BAD_READ) {
	$inputs{$chan}->norm(1);
	return;
	}

# We got a (presumed) valid reading.  Scale it if so configured.
# Multiplying by 1 is pointless, and maybe-might cause floating point
# weirdness, so skip that too.  Then do offset the same way.
$val *= $mul if (defined $mul) and ($mul != 1);
$val += $ofs if (defined $ofs) and ($ofs != 0);

# My car gets 40 rods to the hogshead!
$val = C_to_F($val) if $merica and ($inputs{$chan}->therm);

# store adjusted and normalized values
$inputs{$chan}->cook($val);
if ((defined $min) and (defined $max)) {
	$inputs{$chan}->norm(norm($val, $min, $max));
	}

} # inputs_math_direct

# ----------------------------------------------------------------------
sub inputs_math_agg($chan, $mode) {
# Do math to create an aggregate input.  Depending on $mode, it does it either
# for the cooked values, or for the normalized values (it is called twice,
# one for each, in inputs_math()).

my $intype = $inputs{$chan}->type;

my $agg; # accumulator for aggregation in SUBCHAN loop
my $cnt = 0; # average needs a count
my $win; # remembers highest/lowest channel

# loop through each of the constituent inputs (sub-channels)
# (I don't use "const" in the code because that's confused with "constant")
SUBCHAN: foreach my $subchan (@{$inputs{$chan}->args}) {

	my $val;
	if ($mode == AGG_STRAIGHT) {
		$val = $inputs{$subchan}->cook;
		}
	else {
		$val = $inputs{$subchan}->norm;
		}

	# bad readings are ignored for aggregation purposes
	next SUBCHAN if $val == BAD_READ;

	# for the first valid constituent, just take the value
	if (not defined $agg) {
		$agg = $val;
		$cnt = 1;
		$win = $subchan if $debug;
		next SUBCHAN;
		}

	# second or more, we have to determine what to do
	if    ($intype == IN_MIN) {
		if ($val < $agg) {
			$agg = $val;
			$win = $subchan if $debug;
			}
		}
	elsif ($intype == IN_MAX) {
		if ($val > $agg) {
			$agg = $val;
			$win = $subchan if $debug;
			}
		}
	elsif ($intype == IN_AVG) {
		$agg += $val;
		$cnt++;
		}
	else {
		die 'bogus aggregate input type'; # should never happen
		}

	} # SUBCHAN

if (not defined $agg) {
	# if none of the constituents were any good, we're no good
	$agg = BAD_READ;
	}
elsif ($intype == IN_AVG) {
	# averages need one last step
	$agg /= $cnt;
	}
# min and max don't need further math

dmsg("agg: chan=$chan mode=$mode agg=$agg win=$win", __LINE__) if $debug;

if ($mode == AGG_STRAIGHT) {
	$inputs{$chan}->cook($agg);
	}
else {
	$inputs{$chan}->norm($agg);
	}

} # inputs_math_agg

# ----------------------------------------------------------------------
sub inputs_read($only_slow) {
# Read the various types of sensors.
# If $only_slow is TRUE, we poll slow inputs (and only slow inputs).
# See bounds_main() for why that's useful.	
# If $only_slow is FALSE, we check log_decide().
# If this is a data-log cycle, we poll all inputs, so we can log them.
# If this is *not* a data log cycle, we only poll fast inputs.

CHAN: foreach my $chan (@sinputs) {

	# skip aggregates -- they come from inside this program
	next CHAN if $inputs{$chan}->agg;

	# skip this channel?
	if ($inputs{$chan}->slow) {
		next CHAN unless log_decide() or $only_slow;
		}
	else { # fast
		next CHAN if $only_slow;
		}

	# having decided we want to read it, figure out how
	if ( ($inputs{$chan}->type == IN_FILE)
	or   ($inputs{$chan}->type == IN_PROG) ) {
		read_stream($chan);
		}
	else {
		die 'bad input type'; # should never happen
		}
	}

} # inputs_read

# ----------------------------------------------------------------------
sub read_stream($chan) {
# Read an input sensor value from an I/O stream (which we open here).
# Stream could be a file (such as an hwmon node), or a program/pipe.
# Similar to read_one() but handles special cases, does more checking.

my $path = $inputs{$chan}->path;

dmsg("input: reading $chan", __LINE__) if $debug;

# open the thingy
my ($opened, $infile);
my $intype = $inputs{$chan}->type;
if ($intype eq IN_FILE) {
	$opened = open($infile, READ_FROM, $path);
	}
elsif ($intype eq IN_PROG) {
	my @args = @{$inputs{$chan}->args};
	$opened = open($infile, PIPE_FROM, $path, @args);
	}
else {
	die 'bad input type'; # should never happen
	}
if (not $opened) {
	warn "input: $path: open: $!";
	# don't set a reading, keep BAD_READ
	return;
	}

# read one line from the thingy
my $line = <$infile>;
if (not defined $line) {
	if (eof($infile)) {
		warn "input: $path: read: End of file";
		}
	else {
		warn "input: $path: read: $!";
		}
	return;
	}

# bye-bye thingy
close $infile or warn "input: $path: close: $!";

chomp $line;

# make sure it's a number, then save the reading
$inputs{$chan}->raw($line) if good_num($line, "input '$chan'");

dmsg("input: $chan read as " . $inputs{$chan}->raw, __LINE__) if $debug;

} # read_stream

# ----------------------------------------------------------------------
sub poll_clear() {
# clear previous dynamic values for all input and output channels

foreach my $chan (@sinputs) {
	$inputs{$chan}->raw(BAD_READ);
	$inputs{$chan}->cook(BAD_READ);
	$inputs{$chan}->norm(BAD_READ);
	}

foreach my $out (@outputs) {
	$out->want(undef);
	$out->norm(undef);
	}

} # poll_clear

########################################################################
# data log subroutines

# Writing out sensor data to the data log file.  Not syslog.

# ----------------------------------------------------------------------
sub log_open() {
# Opens the data log (not the system log), if config calls for it.

# need both a log file name and at least one channel to log
return unless $logfile;
return unless @logchans;

open($log, APPEND_TO, $logfile) or die "$logfile: open: $!";

# otherwise log lines sit in buffers indefinitely, possibly being lost
$log->autoflush(1);

# write a header line
log_header();

# log data values with the next poll cycle
log_next();

} # log_open

# ----------------------------------------------------------------------
sub log_reopen() {
# Re-opens the data log file.  Intended for use after log rotation.
# Triggered by $log_reopen flag, which is set in sig_hup() on SIGHUP.

return unless $log;
dmsg('Re-opening log file', __LINE__);
close $log or die "$logfile: close: $!";
log_open();
$log_reopen = FALSE; # clear flag now that we've done it

} # log_reopen

# ----------------------------------------------------------------------
sub log_maybe() {
# Every $logint poll cycles, log data also.
# log_next() can also set $logcyc to force logging on next poll.
# Various routines may do so, such as bounds_main() and sig_usr1().

if (log_decide()) {
	log_data();
	$logcyc = 0;
	}

} # log_maybe

# ----------------------------------------------------------------------
sub log_decide() {
# Returns TRUE if we will be logging data for this poll cycle.
# Called by log_maybe() to determine when to actually log.
# Called by inputs_read() to determine if slow inputs should be read.

return ($logcyc >= $logint);

} # log_decide

# ----------------------------------------------------------------------
sub log_next() {
# set the log cycle counter to the log interval, so we log on next poll

$logcyc = $logint;

} # log_next

# ----------------------------------------------------------------------
sub log_data() {
# logs a line of sensor data, with timestamp
# tab separated, newline at end

return unless $log;

my $msg = EMPTY;

$msg .= $stamp . TAB;
foreach my $chan (@logchans) {
	my $val = get_val($chan);
	$msg .= $val . TAB;
	}
log_write($msg);

} # log_data

# ----------------------------------------------------------------------
sub log_header() {
# print a header line to the log file

# don't need to check $log since we're only called from log_open()

my $msg = EMPTY;

$msg .= 'YYYY-MM-DD HH:MM:SS' . TAB;
foreach my $chan (@logchans) {
	$msg .= $chan . TAB;
	}
log_write($msg);

} # log_header

# ----------------------------------------------------------------------
sub log_write ($msg) {
# write a message to the data log.  log is assumed to be open.

$msg .= NEWLINE;
print {$log} $msg or die "$logfile: write: $!";

} # log_write

########################################################################
# alerting subroutines

# Raise the alarm when a sensor reading is out-of-bounds.

# ----------------------------------------------------------------------
sub bounds_main() {
# Check bounds and times.  Alert if appropriate.
#
# We don't use $poll_repeat for this, as it opens up possibilities for weird
# combinations where something is faulted before and not after.
# Using poll_main(POLL_SLOW_ONLY) is marginally less kludgey.

my $now = time;

# If the last time we warned is within $alert_wait, it was recent.
# $alert_last is initialized to zero, and reset to zero after $reset_wait, so
# in either case, $alerted_recently will end up being false.
my $alerted_recently = ($now - $alert_last) < $alert_wait;

# is anything wrong?
my $faulted = bounds_check();

# If this was not expected to be a log cycle, then slow sensors were not
# read (and will still be at BAD_READ).  Read them, too.  Then check
# again, in case an aggregate contributed to the fault condition.
if ($faulted and not log_decide()) {
	dmsg('checking slow sensors for fault alert', __LINE__);
	log_next();
	poll_main(POLL_SLOW_ONLY);
	$faulted = bounds_check();
	warn('fault cleared after reading slow sensors') unless $faulted;
	}

# remember the last time something was wrong
$fault_last = $now if $faulted;

# If things have been good for at least reset_wait, allow warning again.
# Must come after bounds_check() and $fault_last conditional.
# (If we are actively faulted, $fault_last = $now, diff zero, won't reset.)
$alert_last = 0 if ( ($now - $fault_last) > $reset_wait );

# If there is no current trouble, we're done.
return unless $faulted;

# There's trouble, but we already complained about it recently.  Still done.
if ($alerted_recently) {
	dmsg('alert suppressed due to recent alert', __LINE__);
	return;
	}

# OK, there's trouble, and we haven't complained recently.  Time to whine.

# Thanks to the log_decide() check above, we should have data for all sensors
# (even slow ones), and we are set-up to log it.  The data log timestamp will
# match the $stamp in the alert mail (if any).

# complain
bounds_alert($faulted);

# remember that we complained
$alert_last = $now;

} # bounds_main

# ----------------------------------------------------------------------
sub bounds_check() {
# Check all input values to ensure they are within bounds.
# Return a string with a space-separated list of faulted sensor names,
# with the value in a parenthetical after the name (no space).
# Return an empty string if nothing is wrong.

my $faulted = EMPTY;

CHAN: foreach my $chan (@sinputs) {
	my $val = get_val($chan);
	# if our last reading was bad, don't consider it for checking
	next CHAN if ($val == BAD_READ);
	# compare $val to declared min/max
	my $almin = $inputs{$chan}->almin;
	my $almax = $inputs{$chan}->almax;
	if (  ((defined $almin) and ($val < $almin))
	or    ((defined $almax) and ($val > $almax)) ) {
		# it was out-of-bounds, report it as faulted
		# $chan($val) is not syntax, that's a literal parenthetical
		$faulted .= "$chan($val) ";
		}
	}

return $faulted;

} # bounds_check

# ----------------------------------------------------------------------
sub bounds_alert($faulted) {
# warn my owner that something is wrong

# log to syslog
syslog(LOG_ALERT, "faulted sensors: $faulted");

# broadcast to all users
system(qq{wall 'Hardware sensors faulted: $faulted'}) if $wall;

# email if requested
bounds_mail($faulted) if $mail_to;

} # bounds_alert

# ----------------------------------------------------------------------
sub bounds_mail($faulted) {
# send email about an out-of-bounds condition

my $body = bounds_mail_body($faulted);

# Email::* or IPC::Run seem like overkill here
# just opening a pipe to mail(1) works

my $cmd_mail = qq{mail -s "Hardware sensor fault" $mail_to };
my $opened = open(my $pipe, PIPE_TO, $cmd_mail);
if (not $opened) {
	warn "mail: open: $!";
	return;
	}
print {$pipe} $body or warn "mail: write: $1";
close $pipe or warn "mail: close: $1";

}

# ----------------------------------------------------------------------
sub bounds_mail_body($faulted) {
# return a detailed message body about out-of-bounds sensors

my $all = EMPTY;
foreach my $chan (@sinputs) {
	my $val = get_val($chan);
	$all .= "$chan = $val" . NEWLINE;
	}

# return one giant multi-line string
return <<"EOM";

One or more hardware sensor values are outside declared bounds.

Out of bounds sensors:
$faulted

All current sensor values:
$all

Host: $host
When: $stamp

EOM

} # bounds_mail_body

########################################################################
# configuration subroutines

# Config file, command line, loading, parsing.

# ----------------------------------------------------------------------
sub cfg_init() {
# Initialize runtime configuration.
# Load config from command line and/or files.  Check what we were given to
# make sure it's got everything we need, and is generally sane.  Do any prep
# related to the config, like opening files.

# load from command line or default file
if (@ARGV) {
	# if any command-line arguments were given, use them as config
	cfg_load_argv();
	}
else {
	cfg_load_file(CFGFILE_DEFAULT);
	}

dmsg('config load complete', __LINE__);

cfg_dump() if $debug;

# Even this program needs a purpose in life.
die 'no inputs declared -- nothing to do' unless %inputs;
die 'no output channels, nor data log channels, nor any alarms -- nothing to do'
	unless (@outputs or @logchans or $at_least_one_alm);
cfg_inputs_use_chk();

# math is hard
die 'reset interval must be less than warn interval'
	unless ($reset_wait < $alert_wait);

# log_open() also checks to see if the log should be opened
log_open();

} # cfg_init

# ----------------------------------------------------------------------
sub cfg_dump() {
# Ye Olde "Dump Everything" Routine

dmsg('-------------------- config dump begin');
dmsg("version    = $VERSION");
dmsg("host       = $host");
dmsg("pollint    = <$pollint>");
dmsg("logint     = <$logint>");
dmsg("alert_wait = <$alert_wait>");
dmsg("reset_wait = <$reset_wait>");
dmsg("logfile    = <$logfile>");
dmsg("mail_to    = <$mail_to>");
dmsg("merica     = <$merica>");
dmsg("wall       = <$wall>");
dmsg("quiet      = <$quiet>");
dmsg('-------------------- input channels');
dmsg(join(EMPTY, Dumper(%inputs))) if %inputs;
dmsg('-------------------- sorted inputs');
dmsg(join(SPACE, Dumper(@sinputs))) if @sinputs;
dmsg('-------------------- output channels');
dmsg(join(SPACE, Dumper(@outputs))) if @outputs;
dmsg('-------------------- inits');
dmsg(join(SPACE, Dumper(@inits))) if @inits;
dmsg('-------------------- exits');
dmsg(join(SPACE, Dumper(@exits))) if @exits;
dmsg('-------------------- log channels');
dmsg(join(SPACE, Dumper(@logchans))) if @logchans;
dmsg('-------------------- config dump end');

} # cfg_dump

# ----------------------------------------------------------------------
sub cfg_load_argv() {
# loads config commands from @ARGV array (program command line)

try {
	cfg_load_line(@ARGV);
	}
catch ($e) {
	if (not $cfgerr) {
		emsg("Configuration error on command line\n");
		$cfgerr = 1;
		}
	die $e; # re-throw, we're still dying
	}

} # cfg_load_argv

# ----------------------------------------------------------------------
sub cfg_load_file($cfgfilename) {
# load config commands from a given file

dmsg("loading config file $cfgfilename", __LINE__);

# Some config routines (cfg_path_*) read other files to try and find a sensor
# by name or ID.  That clobbers LAST_FH and related magic.  That means die()
# and warn() don't tell us what line in the config file we were on.  Hence the
# need to track and report line number in here.
my $linenum = 0;

try {

	open(my $cfgfile, READ_FROM, $cfgfilename) or
		die "$cfgfilename: open: $!";

	LINE: while(my $cfgline = <$cfgfile>) {
		$linenum++;
		next LINE if $cfgline =~ m/^\s*$/; # ignore blank lines
		next LINE if $cfgline =~ m/^\s*#/; # ignore comment lines
		my @cfgitems = split SPLIT_ON_WHITESPACE, $cfgline;
		cfg_load_line (@cfgitems);
		}
	}
catch ($e) {
	if (not $cfgerr) {
		my $msg;
		if ($linenum) {
			$msg = "Configuration error at line $linenum of $cfgfilename\n";
			}
		else {
			$msg = "Error loading configuration file $cfgfilename\n";
			}
		emsg($msg);
		$cfgerr = 1;
		}
	die $e; # re-throw, we're still dying
	}

} # cfg_load_file

# ----------------------------------------------------------------------
sub cfg_load_line(@cfgitems) {
# Process a config file command (single line, a list of words).
# The @cfgitems may contain multiple commands.  This splits them up.

my @cmditems = ();
ITEM: while (@cfgitems) {
	my $word = shift @cfgitems;
	if ($word eq SEMI) {
		# word is a semicolon, meaning command separator
		# process what we've got so far
		cfg_load_cmd(@cmditems) if (@cmditems);
		# reset and go around again
		@cmditems = ();
		}
	elsif ($word =~ m/^#/ ) {
		# word begins with pound sign, so rest of line is a comment
		# process what we have so far (after loop exit)
		# ignore the rest
		last ITEM;
		}
	else {
		# accumulate words into the command
		push @cmditems, $word;
		}
	}

# if there were words after the last semicolon, or if there was no semi
cfg_load_cmd(@cmditems) if (@cmditems);

} # cfg_load_line

# ----------------------------------------------------------------------
sub cfg_load_cmd(@cfgitems) {
# Process a config file command (single line, a list of words).
# The @cfgitems list must contain exactly one config command (with any
# arguments), without command separators or comments.

my $cmd = shift @cfgitems;

if ($cmd eq 'debug') {
	$debug = TRUE;
	dmsg('debugging enabled', __LINE__);
	use Data::Dumper;
	}
elsif ($cmd eq 'load') {
	my $filename = $cfgitems[0];
	$filename = CFGFILE_DEFAULT if $filename eq '-';
	cfg_load_file($filename);
	}
elsif (($cmd eq 'merica') or ($cmd eq 'america')) {
	$merica = TRUE;
	}
elsif ($cmd eq 'mail') {
	$mail_to = $cfgitems[0];
	}
elsif ($cmd eq 'wall') {
	$wall = TRUE;
	}
elsif ($cmd eq 'quiet') {
	$quiet = TRUE;
	}
elsif ($cmd eq 'alert') {
	$alert_wait = $cfgitems[0];
	good_int($alert_wait, 'alert-time') or die;
	}
elsif ($cmd eq 'reset') {
	$reset_wait = $cfgitems[0];
	good_int($reset_wait, 'reset-time') or die;
	}
elsif ($cmd eq 'pollint') {
	$pollint = $cfgitems[0];
	good_int($pollint, 'pollint') or die;
	}
elsif ($cmd eq 'logint') {
	$logint = $cfgitems[0];
	good_int($logint, 'logint') or die;
	}
elsif ($cmd eq 'logfile') {
	$logfile = $cfgitems[0];
	}
elsif ($cmd eq 'log') {
	cfg_cmd_logchan(@cfgitems);
	}
elsif ($cmd eq 'input') {
	cfg_cmd_input(@cfgitems);
	}
elsif ($cmd eq 'agg') {
	cfg_cmd_agg(@cfgitems);
	}
elsif ($cmd eq 'output') {
	cfg_cmd_output(@cfgitems);
	}
elsif ($cmd eq 'alarm') {
	cfg_cmd_alm(@cfgitems);
	}
elsif ($cmd eq 'init') {
	cfg_cmd_tweak(\@inits, @cfgitems);
	}
elsif ($cmd eq 'exit') {
	cfg_cmd_tweak(\@exits, @cfgitems);
	}
else {
	die "unrecognized command '$cmd'";
	}

} # cfg_load_cmd

# ----------------------------------------------------------------------
sub cfg_cmd_input (@cfgitems) {
# Process an "input" config command.
# The command "input" has already been stripped from the @cfgitems list.

# input <label> <min> <max> <scale> [opt] <type> <path> [args]
# input CPUt 30 60 0.0001 temp file /sys/class/hwmon/hwmon1/temp2_input

# label/name of input/channel/sensor: must be present, must not already exist
my $label = shift @cfgitems;
die 'input directive missing label' unless defined $label;
die "duplicate input label '$label'" if exists $inputs{$label};

# these three fields are straightforward
my ($min, $max) = (shift @cfgitems, shift @cfgitems);
die 'input directive missing minimum' unless defined $min;
die 'input directive missing maximum' unless defined $max;

# special input bare dash means "no specification for this value"
undef $min if ($min eq '-');
undef $max if ($max eq '-');

# check inputs we were given
# done separately from above for clearer diagnostics (missing vs invalid)
die if defined $min and not good_num($min, 'minimum');
die if defined $max and not good_num($max, 'maximum');
if ( (defined $min) and (defined $max) and (not $min < $max) ) {
	die "input $label: minimum must be less than maximum";
	}

# defaults for options for the INTYPE loop below
# these all correspond to members of the Input structure
my $therm = FALSE;
my $round = FALSE;
my $slow = FALSE;
my $mul = 1; # multiply by one yields the same
my $ofs = 0; # add zero yields the same

# input channel type is required, but might be preceded by options
# loop eating options until we get to a type
my ($arg, $intype, $path, @args);
ARG: while ($arg = shift @cfgitems) {
	# options, will loop again
	if ($arg eq 'temp') {
		$therm = TRUE;
		$mul = MILLIDEGREES;
		$round = TRUE;
		}
	elsif ($arg eq 'round') {
		$round = TRUE;
		}
	elsif ($arg eq 'slow') {
		$slow = TRUE;
		}
	elsif ($arg eq 'mult') {
		$mul = shift @cfgitems;
		die q{input option 'mult' missing value}
			unless defined $mul;
		good_int($mul, 'multiplier') or die;
		}
	elsif ($arg eq 'offset') {
		$ofs = shift @cfgitems;
		die q{input option 'offset' missing value}
			unless defined $ofs;
		good_int($ofs, 'offset') or die;
		}
	# input types, one and only one of these
	elsif ($arg eq 'file') {
		$intype = IN_FILE;
		$path = shift @cfgitems;
		die 'input directive missing path' unless defined $path;
		die 'extra items after file path' if @cfgitems;
		last ARG;
		}
	elsif ($arg eq 'prog') {
		$intype = IN_PROG;
		$path = shift @cfgitems;
		die 'input directive missing program path'
			unless defined $path;
		# any remaining items are arguments to the program
		@args = @cfgitems;
		last ARG;
		}
	elsif ($arg eq 'hwmlabel') {
		$intype = IN_FILE;
		$path = cfg_path_hwmlabel(@cfgitems);
		last ARG;
		}
	elsif ($arg eq 'hwmfile') {
		$intype = IN_FILE;
		$path = cfg_path_hwmfile(@cfgitems);
		last ARG;
		}
	elsif ($arg eq 'wwid') {
		$intype = IN_FILE;
		$path = cfg_path_wwid(@cfgitems);
		last ARG;
		}
	# neither an option nor an input type
	else {
		die "unknown input type/option '$arg'";
		} # if/else ladder
	} # ARG loop

# if the ARG loop exited without finding a type, user left something out
die 'input directive missing input type' unless defined $intype;

# build a structure based on above
my $this = Input->new();
$this->type($intype);
$this->path($path);
$this->min($min);
$this->max($max);
$this->mul($mul);
$this->ofs($ofs);
$this->agg(FALSE);
$this->round($round);
$this->therm($therm);
$this->slow($slow);
$this->args(\@args) if @args;

# now add the struct to the global inputs hash and sinputs list
$inputs{$label} = $this;
push @sinputs, $label;

} # cfg_cmd_input

# ----------------------------------------------------------------------
sub cfg_cmd_agg (@cfgitems) {
# Process an "agg" config command.
# Similar to an "input" command, just with less stuff.	
# The command "agg" has already been stripped from the @cfgitems list.

# agg <label> [opt] <type> <constituents>
# agg MAX round max CPU1 CPU2 SSDa HDDb

# label/name of aggregate channel: must be present, must not already exist
my $label = shift @cfgitems;
die 'aggregate directive missing label' unless defined $label;
die "duplicate aggregate label '$label'" if exists $inputs{$label};

# defaults for options for the ARG loop below
# these all correspond to members of the Input structure
my $round = FALSE;

# aggregate type is required, but might be preceded by options
# loop eating options until we get to a type
my ($arg, $aggtype);
ARG: while ($arg = shift @cfgitems) {
	# options, will loop again
	if ($arg eq 'round') {
		$round = TRUE;
		}
	# aggregate types, one and only one of these
	elsif ($arg eq 'min') {
		$aggtype = IN_MIN;
		last ARG;
		}
	elsif ($arg eq 'max') {
		$aggtype = IN_MAX;
		last ARG;
		}
	elsif ($arg eq 'avg') {
		$aggtype = IN_AVG;
		last ARG;
		}
	# neither an option nor an input type
	else {
		die "unknown aggregate type/option '$arg'";
		} # if/else ladder
	} # ARG loop

# if the ARG loop exited without finding a type, user left something out
die 'aggregate directive missing aggregate type' unless defined $aggtype;

my @args = cfg_agg_chans(@cfgitems);

# build a structure based on above
my $this = Input->new();
$this->type($aggtype);
$this->agg(TRUE);
$this->round($round);
$this->args(\@args);

# now add the struct to the global inputs hash and sinputs list
$inputs{$label} = $this;
push @sinputs, $label;

} # cfg_cmd_agg

# ----------------------------------------------------------------------
sub cfg_cmd_output (@cfgitems) {
# Process an "output" config command.
# The command "output" has already been stripped from the @cfgitems list.

# output <input> <min> <max> <type> <path>
# output FRONT 28 255 file /sys/class/hwmon/hwmon0/pwm2

# ichan/name of controlling input: must be present, not exist, match an input
my $ichan = shift @cfgitems;
die 'output directive missing input parameter' unless defined $ichan;
die "output for unknown input '$ichan'" if not exists $inputs{$ichan};

# these two fields are straightforward
my ($min, $max) = (shift @cfgitems, shift @cfgitems);
die 'output directive missing minimum' unless defined $min;
die 'output directive missing maximum' unless defined $max;

# check inputs we were given
# done separately from above for clearer diagnostics (missing vs invalid)
good_num($min, 'minimum') or die;
good_num($max, 'maximum') or die;

die 'output: minimum must be less than maximum' unless ($min < $max);

# possible options
my $floor;

# output channel type is required, but might be preceded by options
# loop eating options until we get to a type
my ($arg, $path);
ARG: while ($arg = shift @cfgitems) {
	# options, will loop again
	if ($arg eq 'floor') {
		$floor = shift @cfgitems;
		die 'output option floor missing value'
			unless defined $floor;
		good_int($floor, 'floor') or die;
		}
	# output types, one and only one of these
	elsif ($arg eq 'file') {
		$path = shift @cfgitems;
		die 'output directive missing path' unless defined $path;
		die 'extra items after file path' if @cfgitems;
		last ARG;
		}
	elsif ($arg eq 'hwmfile') {
		$path = cfg_path_hwmfile(@cfgitems);
		last ARG;
		}
	# neither an option nor an output type
	else {
		die "unknown output type/option '$arg'";
		}
	}

# if the ARG loop exited without finding a type, user left something out
die 'output directive missing type' unless defined $path;

# build a structure based on the above
my $this = Output->new();
$this->ichan($ichan);
$this->path($path);
$this->min($min);
$this->max($max);
$this->floor($floor) if defined $floor;

# now add the struct to the global outputs list
push @outputs, $this;

# mark the input as used
$inputs{$ichan}->used(TRUE);

} # cfg_cmd_output

# ----------------------------------------------------------------------
sub cfg_cmd_alm (@cfgitems) {
# Process an "alarm" config command.
# The command "alarm" has already been stripped from the @cfgitems list.

# alarm <ichan> <min> <max>
# alarm CPU     0     190

my $ichan = shift @cfgitems;
die 'alarm directive missing input channel name' unless defined $ichan;
die "alarm directive references unknown input channel '$ichan'"
	unless exists $inputs{$ichan};
die "duplicate alarm directive for input channel '$ichan'"
	if defined $inputs{$ichan}->almin;

# get alarm min and max settings
my $almin = shift @cfgitems;
my $almax = shift @cfgitems;

# make sure we got them
die 'alarm directive missing minimum' unless defined $almin;
die 'alarm directive missing maximum' unless defined $almax;

# special input bare dash means "no alarm for this (hi or lo) limit"
undef $almin if ($almin eq '-');
undef $almax if ($almax eq '-');

# make sure what we got is sane
die if defined $almin and not good_num($almin, 'alarm-min');
die if defined $almax and not good_num($almax, 'alarm-max');
if ( (defined $almin) and (defined $almax) and (not $almin < $almax) ) {
	die "alarm $ichan: minimum must be less than maximum";
	}

# save them to the %inputs structure
$inputs{$ichan}->almin($almin);
$inputs{$ichan}->almax($almax);

# make things as used for the config check
$at_least_one_alm = TRUE;
$inputs{$ichan}->used(TRUE);

} # cfg_cmd_alm

# ----------------------------------------------------------------------
sub cfg_cmd_tweak ($tweaklistref, @cfgitems) {
# Process an "init" or "exit" config command, adding the Tweak structure
# to the list passed in $tweaklistref.
# The command name itself has already been stripped from the @cfgitems list.
#
# <command> <value> <type> <path>
# init   1 file /sys/class/hwmon/hwmon0/pwm1_enable
# exit 255 file /sys/class/hwmon/hwmon0/pwm1

my $val = shift @cfgitems;
die 'init/exit directive missing value' unless defined $val;

my $outtype = shift @cfgitems;
die 'init/exit directive missing type' unless defined $outtype;

my $path;
if ($outtype eq 'file') {
	$path = shift @cfgitems;
	die 'init/exit directive missing path' unless defined $path;
	die 'extra items after file path' if @cfgitems;
	}
elsif ($outtype eq 'hwmfile') {
	$path = cfg_path_hwmfile(@cfgitems);
	}
else {
	die "unknown init/exit type '$outtype'" if $outtype ne 'file';
	}

# build a structure based on the above
my $this = Tweak->new();
$this->path($path);
$this->want($val);

# now add the struct to the list of tweaks we were passed
push @{$tweaklistref}, $this;

} # cfg_cmd_tweak

# ----------------------------------------------------------------------
sub cfg_cmd_logchan (@cfgitems) {
# Process a "log" config command.
# The command name has already been stripped from the @cfgitems list.

# log <ichan> [ichan...]
# log CPU GPU HDD

# make sure we were given at least one channel
die 'log directive missing channel(s)' unless @cfgitems;

# make sure the channel(s) given exist as inputs
foreach my $chan (@cfgitems) {
	die "log directive references unknown input channel '$chan'"
		unless exists $inputs{$chan};
	# mark the input as used
	$inputs{$chan}->used(TRUE);
	}

@logchans = @cfgitems;

} # cfg_cmd_logchan

# ----------------------------------------------------------------------
sub cfg_agg_chans(@cfgitems) {
# Load a load a list of input channels.
# Basically just a validation, then we return what we were given.

die 'aggregate input missing constituent inputs' unless @cfgitems;

# Make sure each constituent input was already declared.
ITEM: foreach my $item (@cfgitems) {
	if (exists $inputs{$item}) {
		$inputs{$item}->used(TRUE);
		}
	else {
		die "unknown constituent input '$item'";
		}
	}

return @cfgitems;

} # cfg_agg_chans

# ----------------------------------------------------------------------
sub cfg_path_hwmlabel (@cfgitems) {
# Return the path to a hwmon node, given its device name and channel label in
# @cfgitems.  Abort program if extra @cfgitems or not found.

my $devname = shift @cfgitems or die 'hwmlabel missing device name';
my $label   = shift @cfgitems or die 'hwmlabel missing channel label';
die 'extra items after hwmon channel label' if @cfgitems;

dmsg("hwmlabel: looking for hwmon <$devname> label <$label>", __LINE__);

my $match; # path to hwmon label node, if we find one

# look for a hwmon device that matches $devname
DEV: while (my $devpath = glob('/sys/class/hwmon/hwmon*/name')) {
	my $maybe = read_one($devpath);
	dmsg("hwmlabel: $devpath = <$maybe>", __LINE__);
	next DEV unless $maybe;
	next DEV unless $maybe eq $devname;
	$devpath = dirname($devpath);
	# look for a hwmon node under that, that matches $label
	CHAN: while (my $labelpath = glob("$devpath/*_label")) {
		$maybe = read_one($labelpath);
		dmsg("hwmlabel: $labelpath = <$maybe>", __LINE__);
		next CHAN unless $maybe;
		next CHAN unless $maybe eq $label;
		die "multiple matches for hwmon device '$devname', label '$label'"
			if defined $match;
		$match = $labelpath;
		# After a match, we continue through the rest.
		# If more than one matches, that's as bad as no matches.
		} # CHAN
	} # DEV

die "no matches for hwmon device '$devname', label '$label'"
	unless defined $match;

# replace the _label we found with the _input that we'll read
$match =~ s/_label$/_input/ or
	die 'unexpected regex failure'; # should never happen

# make sure there actually *is* an input to read
-e $match or die "$match: does not exist ($devname:$label)";

dmsg("hwmlabel: found <$match>", __LINE__);

return $match;

} # cfg_path_hwmlabel

# ----------------------------------------------------------------------
sub cfg_path_hwmfile (@cfgitems) {
# Return the path to a hwmon node, given its device name and channel file name
# in @cfgitems.  Abort program if extra @cfgitems or not found.

my $devname  = shift @cfgitems or die 'hwmfile missing device name';
my $chanfile = shift @cfgitems or die 'hwmfile missing file name';
die 'extra items after hwmon file name' if @cfgitems;

dmsg("hwmfile: looking for hwmon <$devname> file <$chanfile>", __LINE__);

my $match; # path to the channel file, if we find one

# look for a hwmon device that matches $devname
DEV: while (my $path = glob('/sys/class/hwmon/hwmon*/name')) {
	my $maybe = read_one($path);
	dmsg("hwmfile: $path = <$maybe>", __LINE__);
	next DEV unless $maybe;
	next DEV unless $maybe eq $devname;
	$path = dirname($path);
	$path = "$path/$chanfile";
	# look for a hwmon node under that, that matches given filename
	if (-e $path) {
		die "multiple matches for hwmon device '$devname', file '$chanfile'"
			if defined $match;
		$match = $path;
		# After a match, we continue through the rest.
		# If more than one matches, that's as bad as no matches.
		}
	} # DEV

die "no matches for hwmon device '$devname', file '$chanfile'"
	unless defined $match;

dmsg("hwmfile: found <$match>", __LINE__);

return $match;

} # cfg_path_hwmfile

# ----------------------------------------------------------------------
sub cfg_path_wwid (@cfgitems) {
# Return the path to a hwmon node, assuming the hwmon device is the drivetemp
# module, and given its WWID (worldwide name) as the sole element of
# @cfgitems.  Abort program if extra @cfgitems or not found.

my $wwid = shift @cfgitems or die 'wwid directive missing WWID';
die 'extra items after WWID' if @cfgitems;

dmsg("wwid: looking for WWID <$wwid>", __LINE__);

my $match; # path to the matching WWID file, if we find one

# look for a hwmon device that matches module "drivetemp"
DEV: while (my $path = glob('/sys/class/hwmon/hwmon*/name')) {
	my $maybe = read_one($path);
	dmsg("wwid: dev : $path = <$maybe>", __LINE__);
	next DEV unless $maybe;
	next DEV unless $maybe eq 'drivetemp';
	# having found a drivetemp entry, see if it has the WWID we want
	$path = dirname($path);
	$path = "$path/device/wwid";
	$maybe = read_one($path);
	next DEV unless $maybe;
	dmsg("wwid: node: $path = <$maybe>", __LINE__);
	next DEV unless $maybe eq $wwid;
	die "multiple matches for WWID '$wwid'"	if defined $match;
	$match = $path;
	# After a match, we continue through the rest.
	# If more than one matches, that's as bad as no matches.
	} # DEV

die "no matches for WWID '$wwid'" unless defined $match;

# replace the WWID node we found with the temp1_input that we'll read
$match =~ s{device/wwid}{temp1_input} or
	die 'unexpected regex failure'; # should never happen

# make sure there actually *is* an input to read
-e $match or die "$match: does not exist (WWID $wwid)";

dmsg("wwid: found <$match>", __LINE__);

return $match;

} # cfg_path_wwid

# ----------------------------------------------------------------------
sub cfg_inputs_use_chk() {
# checks each input to make sure it is used by something

foreach my $ichan (@sinputs) {
	warn "WARNING: input '$ichan' is not used by any output, alarm, or log"
		unless $inputs{$ichan}->used;
	}

} # cfg_inputs_use_chk

# ----------------------------------------------------------------------
sub read_one($path) {
# Reads one line of data from a given file name.
# On failure, writes warnings, returns undef.
# Similar to inputs_read() but more general, does fewer specialty steps.

my ($opened, $infile);
$opened = open($infile, READ_FROM, $path);
if (not $opened) {
	warn "$path: open: $!";
	return;
	}

my $line = <$infile>;
if (not defined $line) {
	if (eof($infile)) {
		warn " $path: read: End of file";
		}
	else {
		warn "$path: read: $!";
		}
	return;
	}

close $infile or warn "$path: close: $!";

chomp $line;

return $line;

} # read_one

########################################################################
# miscellaneous/utility subroutines

# ----------------------------------------------------------------------
sub emsg ($msg) {
# error message - to stderr and syslog
# You need to bring your own newline.

syslog(LOG_ERR, $msg);
$msg = "$PROGNAME: $msg";
print {*STDERR} $msg;

} # emsg

# ----------------------------------------------------------------------
sub dmsg ($msg, $line = undef) {
# debug message
# Do nothing if not debugging.
# Write the message to syslog.
# Also write same to stderr, unless configured not to with "$quiet"

# For some calls, I put a conditional on the call, too, to avoid
# overhead from string interpolation, etc., in every loop.
return unless $debug;

$msg = "DEBUG: $msg";
$msg = "$line: $msg" if defined $line;
syslog(LOG_DEBUG, $msg);

return if $quiet;

$msg = "$PROGNAME: $msg";
say {*STDERR} $msg;

} # dmsg

# ----------------------------------------------------------------------
sub get_val ($chan) {
# Returns the last value (reading) for an input sensor.
# Simply a convenience wrapper.

return $inputs{$chan}->cook;

} # get_val

# ----------------------------------------------------------------------
sub good_num ($in, $name) {
# Tests $in to see if it is a number acceptable to us and Perl.
# Decimal fractions and negative values are good, infinity is not.
# If all tests pass, returns true.
# If any test fails, warns, using $name to specify, and returns false.

my $re_num = qr{^[-+.\d]+$}; # characters valid in any real number

my $good = FALSE;

if ($in !~ $re_num) {
	warn "$name contains a non-numeric character";
	}
elsif (not looks_like_number($in)) {
	warn "$name does not look like a number";
	}
else {
	$good = TRUE;
	}

return $good;

} # good_num

# ----------------------------------------------------------------------
sub good_int ($in, $name) {
# Tests $in to see if it is a positive integer.
# If all tests pass, returns true.
# If any test fails, warns, using $name to specify, and returns false.

# Who's a good int?  Who's a good little int?  Is it you?  It is you?

my $re_digits = qr{^\d+$}; # one-or-more digits

my $good = FALSE;

if ($in !~ $re_digits) {
	warn "$name contains a non-numeric character";
	}
elsif (not looks_like_number($in)) {
	warn "$name does not look like a number";
	}
elsif ($in < 1) {
	warn "$name is less than one";
	}
else {
	$good = TRUE;
	}

return $good;

} # good_int

# ----------------------------------------------------------------------
sub norm ($act, $min, $max) {
# Computes a normalized value for an actual value.
# A normalized value has a value from 0 to 1, inclusive, proportional to the
# domain (possible input values), with 1 being the highest point of the
# domain, and 0 being the lowest point of the domain.

return ($act - $min) / ($max - $min);

} # norm

# ----------------------------------------------------------------------
sub denorm ($norm, $min, $max) {
# Computes an actual value from a normalized value, i.e., denormalizes.
# Rounds the result up to the next-highest integer as needed.

return ceil($norm * ( $max - $min ) + $min);

} # denorm

# ----------------------------------------------------------------------
sub C_to_F ($C) {
# Celsius to Fahrenheit

return unless defined $C;
my $F = ((9/5) * $C) + 32;
return $F;

} # C_to_F

## END #################################################################
